@@ -1,24 +1,3 @@
-# ------------------------------------------------------------------------------------------------ #
-# MIT License                                                                                      #
-#                                                                                                  #
-# Copyright (c) 2020, Microsoft Corporation                                                        #
-#                                                                                                  #
-# Permission is hereby granted, free of charge, to any person obtaining a copy of this software    #
-# and associated documentation files (the "Software"), to deal in the Software without             #
-# restriction, including without limitation the rights to use, copy, modify, merge, publish,       #
-# distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the    #
-# Software is furnished to do so, subject to the following conditions:                             #
-#                                                                                                  #
-# The above copyright notice and this permission notice shall be included in all copies or         #
-# substantial portions of the Software.                                                            #
-#                                                                                                  #
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING    #
-# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND       #
-# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,     #
-# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,   #
-# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.          #
-# ------------------------------------------------------------------------------------------------ #
-
 import gc
 import unittest
 from collections import namedtuple
@@ -137,6 +116,7 @@ class TestCase(unittest.TestCase):
                 safe_sample(self.env_discrete.action_space, seed=(self.seed * i * 13))
                 for i in range(1, 12)], axis=0),
             logP_next=onp.log(onp.random.RandomState(17).rand(11)),
+            extra_info=None
         )
 
     @property
@@ -160,6 +140,7 @@ class TestCase(unittest.TestCase):
                 safe_sample(self.env_boxspace.action_space, seed=(self.seed * i * 13))
                 for i in range(1, 12)], axis=0),
             logP_next=onp.log(onp.random.RandomState(17).rand(11)),
+            extra_info=None
         )
 
     @property
@@ -230,7 +211,7 @@ class TestCase(unittest.TestCase):
         return func
 
     @property
-    def func_q_tochastic_type1(self):
+    def func_q_stochastic_type1(self):
         def func(S, A, is_training):
             flatten = hk.Flatten()
             batch_norm = hk.BatchNorm(create_scale=True, create_offset=True, decay_rate=0.95)
@@ -247,7 +228,7 @@ class TestCase(unittest.TestCase):
         return func
 
     @property
-    def func_q_tochastic_type2(self):
+    def func_q_stochastic_type2(self):
         def func(S, is_training):
             flatten = hk.Flatten()
             batch_norm = hk.BatchNorm(create_scale=True, create_offset=True, decay_rate=0.95)
@@ -258,7 +239,7 @@ class TestCase(unittest.TestCase):
                 hk.Linear(self.env_discrete.action_space.n * 51),
                 hk.Reshape((self.env_discrete.action_space.n, 51))
             ))
-            return seq(flatten(S))
+            return {'logits': seq(flatten(S))}
         return func
 
     @property
