@@ -1,24 +1,3 @@
-# ------------------------------------------------------------------------------------------------ #
-# MIT License                                                                                      #
-#                                                                                                  #
-# Copyright (c) 2020, Microsoft Corporation                                                        #
-#                                                                                                  #
-# Permission is hereby granted, free of charge, to any person obtaining a copy of this software    #
-# and associated documentation files (the "Software"), to deal in the Software without             #
-# restriction, including without limitation the rights to use, copy, modify, merge, publish,       #
-# distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the    #
-# Software is furnished to do so, subject to the following conditions:                             #
-#                                                                                                  #
-# The above copyright notice and this permission notice shall be included in all copies or         #
-# substantial portions of the Software.                                                            #
-#                                                                                                  #
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING    #
-# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND       #
-# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,     #
-# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,   #
-# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.          #
-# ------------------------------------------------------------------------------------------------ #
-
 import jax
 import numpy as onp
 import chex
@@ -52,10 +31,10 @@ class PrioritizedReplayBuffer(BaseReplayBuffer):
     TD errors collected from a value-function updater, e.g. :func:`QLearning.td_error
     <coax.td_learning.QLearning.td_error>`.
 
-    Since the prioritized samples are baised, the :attr:`sample` method also produces non-trivial
+    Since the prioritized samples are biased, the :attr:`sample` method also produces non-trivial
     importance weights (stored in the :class:`TransitionBatch.W
     <coax.reward_tracing.TransitionBatch>` attribute). The logic for constructing these weights for
-    a sample of batch size :math:`n` (typically :math:`n\ll N`):
+    a sample of batch size :math:`n` is:
 
     .. math::
 
