@@ -1,24 +1,3 @@
-# ------------------------------------------------------------------------------------------------ #
-# MIT License                                                                                      #
-#                                                                                                  #
-# Copyright (c) 2020, Microsoft Corporation                                                        #
-#                                                                                                  #
-# Permission is hereby granted, free of charge, to any person obtaining a copy of this software    #
-# and associated documentation files (the "Software"), to deal in the Software without             #
-# restriction, including without limitation the rights to use, copy, modify, merge, publish,       #
-# distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the    #
-# Software is furnished to do so, subject to the following conditions:                             #
-#                                                                                                  #
-# The above copyright notice and this permission notice shall be included in all copies or         #
-# substantial portions of the Software.                                                            #
-#                                                                                                  #
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING    #
-# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND       #
-# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,     #
-# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,   #
-# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.          #
-# ------------------------------------------------------------------------------------------------ #
-
 from abc import ABC, abstractmethod
 
 import jax
@@ -29,13 +8,15 @@ import chex
 
 from .._base.mixins import RandomStateMixin
 from ..utils import get_grads_diagnostics, is_policy, is_stochastic, is_qfunction, is_vfunction, jit
-from ..value_losses import huber
+from ..value_losses import huber, quantile_huber
 from ..regularizers import Regularizer
+from ..proba_dists import DiscretizedIntervalDist, EmpiricalQuantileDist
 
 
 __all__ = (
     'BaseTDLearningV',
     'BaseTDLearningQ',
+    'BaseTDLearningQWithTargetPolicy'
 )
 
 
@@ -106,10 +87,10 @@ class BaseTDLearning(ABC, RandomStateMixin):
         grads, function_state, metrics, td_error = self.grads_and_metrics(transition_batch)
         if any(jnp.any(jnp.isnan(g)) for g in jax.tree_leaves(grads)):
             raise RuntimeError(f"found nan's in grads: {grads}")
-        self.update_from_grads(grads, function_state)
+        self.apply_grads(grads, function_state)
         return (metrics, td_error) if return_td_error else metrics
 
-    def update_from_grads(self, grads, function_state):
+    def apply_grads(self, grads, function_state):
         r"""
 
         Update the model parameters (weights) of the underlying function approximator given
@@ -255,14 +236,17 @@ class BaseTDLearningV(BaseTDLearning):
             S = self.v.observation_preprocessor(next(rngs), transition_batch.S)
             W = jnp.clip(transition_batch.W, 0.1, 10.)  # clip importance weights to reduce variance
 
+            metrics = {}
+
             # regularization term
             if self.policy_regularizer is None:
                 regularizer = 0.
             else:
-                # flip sign (typical example: regularizer = -beta * entropy)
-                regularizer = -self.policy_regularizer.batch_eval(
+                regularizer, regularizer_metrics = self.policy_regularizer.batch_eval(
                     target_params['reg'], target_params['reg_hparams'], target_state['reg'],
                     next(rngs), transition_batch)
+                metrics.update({f'{self.__class__.__name__}/{k}': v for k,
+                               v in regularizer_metrics.items()})
 
             if is_stochastic(self.v):
                 dist_params, state_new = self.v.function(params, state, next(rngs), S, True)
@@ -271,9 +255,15 @@ class BaseTDLearningV(BaseTDLearning):
 
                 if self.policy_regularizer is not None:
                     dist_params_target = self.v.proba_dist.affine_transform(
-                        dist_params_target, 1., regularizer, self.v.value_transform)
+                        dist_params_target, 1., -regularizer, self.v.value_transform)
 
-                loss = jnp.mean(self.v.proba_dist.cross_entropy(dist_params_target, dist_params))
+                if isinstance(self.v.proba_dist, DiscretizedIntervalDist):
+                    loss = jnp.mean(self.v.proba_dist.cross_entropy(dist_params_target,
+                                                                    dist_params))
+                elif isinstance(self.v.proba_dist, EmpiricalQuantileDist):
+                    loss = quantile_huber(dist_params_target['values'],
+                                          dist_params['values'],
+                                          dist_params['quantile_fractions'], W)
 
                 # the rest here is only needed for metrics dict
                 V = self.v.proba_dist.mean(dist_params)
@@ -288,7 +278,8 @@ class BaseTDLearningV(BaseTDLearning):
             else:
                 V, state_new = self.v.function(params, state, next(rngs), S, True)
                 G = self.target_func(target_params, target_state, next(rngs), transition_batch)
-                G += regularizer
+                # flip sign (typical example: regularizer = -beta * entropy)
+                G -= regularizer
                 loss = self.loss_function(G, V, W)
 
                 # only needed for metrics dict
@@ -300,11 +291,11 @@ class BaseTDLearningV(BaseTDLearning):
             dLoss_dV = jax.grad(self.loss_function, argnums=1)
             td_error = -V.shape[0] * dLoss_dV(G, V)  # e.g. (G - V) if loss function is MSE
             chex.assert_equal_shape([td_error, W])
-            metrics = {
+            metrics.update({
                 f'{self.__class__.__name__}/loss': loss,
                 f'{self.__class__.__name__}/td_error': jnp.mean(W * td_error),
                 f'{self.__class__.__name__}/td_error_targ': jnp.mean(-dLoss_dV(V, V_targ, W)),
-            }
+            })
             return loss, (td_error, state_new, metrics)
 
         def grads_and_metrics_func(
@@ -401,14 +392,17 @@ class BaseTDLearningQ(BaseTDLearning):
             A = self.q.action_preprocessor(next(rngs), transition_batch.A)
             W = jnp.clip(transition_batch.W, 0.1, 10.)  # clip importance weights to reduce variance
 
+            metrics = {}
+
             # regularization term
             if self.policy_regularizer is None:
                 regularizer = 0.
             else:
-                # flip sign (typical example: regularizer = -beta * entropy)
-                regularizer = -self.policy_regularizer.batch_eval(
+                regularizer, regularizer_metrics = self.policy_regularizer.batch_eval(
                     target_params['reg'], target_params['reg_hparams'], target_state['reg'],
                     next(rngs), transition_batch)
+                metrics.update({f'{self.__class__.__name__}/{k}': v for k,
+                               v in regularizer_metrics.items()})
 
             if is_stochastic(self.q):
                 dist_params, state_new = \
@@ -418,9 +412,15 @@ class BaseTDLearningQ(BaseTDLearning):
 
                 if self.policy_regularizer is not None:
                     dist_params_target = self.q.proba_dist.affine_transform(
-                        dist_params_target, 1., regularizer, self.q.value_transform)
+                        dist_params_target, 1., -regularizer, self.q.value_transform)
 
-                loss = jnp.mean(self.q.proba_dist.cross_entropy(dist_params_target, dist_params))
+                if isinstance(self.q.proba_dist, DiscretizedIntervalDist):
+                    loss = jnp.mean(self.q.proba_dist.cross_entropy(dist_params_target,
+                                                                    dist_params))
+                elif isinstance(self.q.proba_dist, EmpiricalQuantileDist):
+                    loss = quantile_huber(dist_params_target['values'],
+                                          dist_params['values'],
+                                          dist_params['quantile_fractions'], W)
 
                 # the rest here is only needed for metrics dict
                 Q = self.q.proba_dist.mean(dist_params)
@@ -435,7 +435,8 @@ class BaseTDLearningQ(BaseTDLearning):
             else:
                 Q, state_new = self.q.function_type1(params, state, next(rngs), S, A, True)
                 G = self.target_func(target_params, target_state, next(rngs), transition_batch)
-                G += regularizer
+                # flip sign (typical example: regularizer = -beta * entropy)
+                G -= regularizer
                 loss = self.loss_function(G, Q, W)
 
                 # only needed for metrics dict
@@ -447,11 +448,11 @@ class BaseTDLearningQ(BaseTDLearning):
             dLoss_dQ = jax.grad(self.loss_function, argnums=1)
             td_error = -Q.shape[0] * dLoss_dQ(G, Q)  # e.g. (G - Q) if loss function is MSE
             chex.assert_equal_shape([td_error, W])
-            metrics = {
+            metrics.update({
                 f'{self.__class__.__name__}/loss': loss,
                 f'{self.__class__.__name__}/td_error': jnp.mean(W * td_error),
                 f'{self.__class__.__name__}/td_error_targ': jnp.mean(-dLoss_dQ(Q, Q_targ, W)),
-            }
+            })
             return loss, (td_error, state_new, metrics)
 
         def grads_and_metrics_func(
