@@ -1,26 +1,7 @@
-# ------------------------------------------------------------------------------------------------ #
-# MIT License                                                                                      #
-#                                                                                                  #
-# Copyright (c) 2020, Microsoft Corporation                                                        #
-#                                                                                                  #
-# Permission is hereby granted, free of charge, to any person obtaining a copy of this software    #
-# and associated documentation files (the "Software"), to deal in the Software without             #
-# restriction, including without limitation the rights to use, copy, modify, merge, publish,       #
-# distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the    #
-# Software is furnished to do so, subject to the following conditions:                             #
-#                                                                                                  #
-# The above copyright notice and this permission notice shall be included in all copies or         #
-# substantial portions of the Software.                                                            #
-#                                                                                                  #
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING    #
-# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND       #
-# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,     #
-# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,   #
-# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.          #
-# ------------------------------------------------------------------------------------------------ #
-
 import warnings
+from functools import partial
 
+import chex
 import gym
 import jax
 import jax.numpy as jnp
@@ -50,7 +31,10 @@ __all__ = (
     'merge_dicts',
     'single_to_batch',
     'safe_sample',
+    'stack_trees',
     'tree_ravel',
+    'tree_sample',
+    'unvectorize',
 )
 
 
@@ -153,7 +137,7 @@ def batch_to_single(pytree, index=0):
 
 def check_array(
         arr, ndim=None, ndim_min=None, ndim_max=None,
-        dtype=None, shape=None, axis_size=None, axis=None):
+        dtype=None, shape=None, axis_size=None, axis=None, except_np=False):
     r"""
 
     This helper function is mostly for internal use. It is used to check a few
@@ -166,8 +150,10 @@ def check_array(
         If one of the checks fails.
 
     """
-    if not isinstance(arr, jnp.ndarray):
-        raise TypeError("expected input to be a numpy array, got type: {}".format(type(arr)))
+    if not except_np and not isinstance(arr, jnp.ndarray):
+        raise TypeError(f"expected input to be a jnp.ndarray, got type: {type(arr)}")
+    if not isinstance(arr, (onp.ndarray, jnp.ndarray)):
+        raise TypeError(f"expected input to be an ndarray, got type: {type(arr)}")
 
     check = ndim is not None
     ndims = [ndim] if not isinstance(ndim, (list, tuple, set)) else ndim
@@ -284,9 +270,8 @@ def chunks_pow2(transition_batch):
 def clipped_logit(x, epsilon=1e-15):
     r"""
 
-    A safe implementation of the logit function
-    :math:`x\mapsto\log(x/(1-x))`. It clips the arguments of the log function
-    from below so as to avoid evaluating it at 0:
+    A safe implementation of the logit function :math:`x\mapsto\log(x/(1-x))`. It clips the
+    arguments of the log function from below so as to avoid evaluating it at 0:
 
     .. math::
 
@@ -297,13 +282,11 @@ def clipped_logit(x, epsilon=1e-15):
     ----------
     x : ndarray
 
-        Input numpy array whose entries lie on the unit interval,
-        :math:`x_i\in [0, 1]`.
+        Input numpy array whose entries lie on the unit interval, :math:`x_i\in [0, 1]`.
 
     epsilon : float, optional
 
-        The small number with which to clip the arguments of the logarithm from
-        below.
+        The small number with which to clip the arguments of the logarithm from below.
 
     Returns
     -------
@@ -313,9 +296,7 @@ def clipped_logit(x, epsilon=1e-15):
         :math:`z_i\in\mathbb{R}`.
 
     """
-    if jax.api._jit_is_disabled():
-        assert jnp.all(x >= 0) and jnp.all(x <= 1), "values do not lie on the unit interval"
-    return jnp.log(jnp.maximum(epsilon, x)) - jnp.log(jnp.maximum(epsilon, 1 - x))
+    return jnp.log(jnp.clip(x, epsilon, 1)) - jnp.log(jnp.clip(1 - x, epsilon, 1))
 
 
 def default_preprocessor(space):
@@ -356,10 +337,13 @@ def default_preprocessor(space):
 
     elif isinstance(space, gym.spaces.MultiDiscrete):
         def func(rng, X):
-            rngs = hk.PRNGSequence(rng)
+            rngs = jax.random.split(rng, len(space.nvec))
+            chex.assert_rank(X, {1, 2})
+            if X.ndim == 1:
+                X = jnp.expand_dims(X, axis=0)
             return [
-                default_preprocessor(gym.spaces.Discrete(n))(next(rngs), X[i])
-                for i, n in enumerate(space.nvec)]
+                default_preprocessor(gym.spaces.Discrete(n))(rng, X[:, i])
+                for i, (n, rng) in enumerate(zip(space.nvec, rngs))]
 
     elif isinstance(space, gym.spaces.MultiBinary):
         def func(rng, X):
@@ -769,6 +753,7 @@ class StepwiseLinearFunction:
 
 
     """
+
     def __init__(self, *steps):
         if len(steps) < 2:
             raise TypeError("need at least two steps")
@@ -904,3 +889,186 @@ def tree_ravel(pytree):
 
     """
     return jnp.concatenate([jnp.ravel(leaf) for leaf in jax.tree_leaves(pytree)])
+
+
+def tree_sample(pytree, rng, n=1, replace=False, axis=0, p=None):
+    r"""
+
+    Flatten and concatenate all leaves into a single flat ndarray.
+
+    Parameters
+    ----------
+    pytree : a pytree with ndarray leaves
+
+        A typical example is a pytree of model parameters (weights) or gradients with respect to
+        such model params.
+
+    rng : jax.random.PRNGKey
+
+        A pseudo-random number generator key.
+
+    n : int, optional
+
+        The sample size. Note that the smaple size cannot exceed the batch size of the provided
+        :code:`pytree` if :code:`with_replacement=False`.
+
+    replace : bool, optional
+
+        Whether to sample with replacement.
+
+    axis : int, optional
+
+        The axis along which to sample.
+
+    p : 1d array, optional
+
+        The sampling propensities.
+
+    Returns
+    -------
+    arr : ndarray with ndim=1
+
+        A single flat array.
+
+    """
+    batch_size = _check_leaf_batch_size(pytree)
+    ix = jax.random.choice(rng, batch_size, shape=(n,), replace=replace, p=p)
+    return jax.tree_map(lambda x: jnp.take(x, ix, axis=0), pytree)
+
+
+def unvectorize(f, in_axes=0, out_axes=0):
+    """
+
+    Apply a batched function on a single instance, which effectively does the
+    inverse of what :func:`jax.vmap` does.
+
+    Parameters
+    ----------
+    f : callable
+
+        A batched function.
+
+    in_axes : int or tuple of ints, optional
+
+        Specify the batch axes of the inputs of the function :code:`f`. If left unpsecified, this
+        defaults to axis 0 for all inputs.
+
+    out_axis: int, optional
+
+        Specify the batch axes of the outputs of the function :code:`f`. These axes will be dropped
+        by :func:`jnp.squeeze <jax.numpy.squeeze>`, i.e. dropped. If left unpsecified, this defaults
+        to axis 0 for all outputs.
+
+    Returns
+    -------
+    f_single : callable
+
+        The unvectorized version of :code:`f`.
+
+    Examples
+    --------
+
+    Haiku uses a batch-oriented design (although some components may be batch-agnostic). To create a
+    function that acts on a single instance, we can use :func:`unvectorize` as follows:
+
+    .. code:: python
+
+        import jax.numpy as jnp
+        import haiku as hk
+        import coax
+
+
+        def f(x_batch):
+            return hk.Linear(11)(x_batch)
+
+
+        rngs = hk.PRNGSequence(42)
+
+        x_batch = jnp.zeros(shape=(3, 5))  # batch of 3 instances
+        x_single = jnp.zeros(shape=(5,))   # single instance
+
+        init, f_batch = hk.transform(f)
+        params = init(next(rngs), x_batch)
+        y_batch = f_batch(params, next(rngs), x_batch)
+        assert y_batch.shape == (3, 11)
+
+        f_single = coax.unvectorize(f_batch, in_axes=(None, None, 0), out_axes=0)
+        y_single = f_single(params, next(rngs), x_single)
+        assert y_single.shape == (11,)
+
+    Alternatively, and perhaps more conveniently, we can unvectorize the function before doing the
+    Haiku transform:
+
+    .. code:: python
+
+        init, f_single = hk.transform(coax.unvectorize(f))
+        params = init(next(rngs), x_single)
+        y_single = f_single(params, next(rngs), x_single)
+        assert y_single.shape == (11,)
+
+
+    """
+    def f_single(*args):
+        in_axes_ = in_axes
+        if in_axes is None or isinstance(in_axes, int):
+            in_axes_ = (in_axes,) * len(args)
+        if len(args) != len(in_axes_):
+            raise ValueError("number of in_axes must match the number of function inputs")
+        vargs = [
+            arg if axis is None else
+            jax.tree_map(partial(jnp.expand_dims, axis=axis), arg)
+            for arg, axis in zip(args, in_axes_)]
+        out = f(*vargs)
+        out_axes_ = out_axes
+        if isinstance(out, tuple):
+            if out_axes_ is None or isinstance(out_axes_, int):
+                out_axes_ = (out_axes_,) * len(out)
+            if len(out) != len(out_axes_):
+                raise ValueError("number of out_axes must match the number of function outputs")
+            out = tuple(
+                x if axis is None else
+                jax.tree_map(partial(jnp.squeeze, axis=axis), x)
+                for x, axis in zip(out, out_axes_))
+        elif out_axes_ is not None:
+            if not isinstance(out_axes_, int):
+                raise TypeError(
+                    "out_axes must be an int for functions with a single output; "
+                    f"got: out_axes={out_axes}")
+            out = jax.tree_map(partial(jnp.squeeze, axis=out_axes), out)
+        return out
+    return f_single
+
+
+def _check_leaf_batch_size(pytree):
+    """ some boilerplate to extract the batch size with some consistency checks """
+    leaf, *leaves = jax.tree_leaves(pytree)
+    if not isinstance(leaf, (onp.ndarray, jnp.ndarray)) and leaf.ndim >= 1:
+        raise TypeError(f"all leaves must be arrays; got type: {type(leaf)}")
+    if leaf.ndim < 1:
+        raise TypeError("all leaves must be at least 1d, i.e. (batch_size, ...)")
+    batch_size = leaf.shape[0]
+    for leaf in leaves:
+        if not isinstance(leaf, (onp.ndarray, jnp.ndarray)) and leaf.ndim >= 1:
+            raise TypeError(f"all leaves must be arrays; got type: {type(leaf)}")
+        if leaf.ndim < 1:
+            raise TypeError("all leaves must be at least 1d, i.e. (batch_size, ...)")
+        if leaf.shape[0] != batch_size:
+            raise TypeError("all leaves must have the same batch_size")
+    return batch_size
+
+
+def stack_trees(*trees):
+    """
+    Stack
+    Parameters
+    ----------
+    trees : sequence of pytrees with ndarray leaves
+        A typical example are pytrees containing the parameters and function states of
+        a model that should be used in a function which is vectorized by `jax.vmap`. The trees
+        have to have the same pytree structure.
+    Returns
+    -------
+    pytree : pytree with ndarray leaves
+        A tuple of pytrees.
+    """
+    return jax.tree_util.tree_multimap(lambda *args: jnp.stack(args), *zip(*trees))
