@@ -1,30 +1,8 @@
-# ------------------------------------------------------------------------------------------------ #
-# MIT License                                                                                      #
-#                                                                                                  #
-# Copyright (c) 2020, Microsoft Corporation                                                        #
-#                                                                                                  #
-# Permission is hereby granted, free of charge, to any person obtaining a copy of this software    #
-# and associated documentation files (the "Software"), to deal in the Software without             #
-# restriction, including without limitation the rights to use, copy, modify, merge, publish,       #
-# distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the    #
-# Software is furnished to do so, subject to the following conditions:                             #
-#                                                                                                  #
-# The above copyright notice and this permission notice shall be included in all copies or         #
-# substantial portions of the Software.                                                            #
-#                                                                                                  #
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING    #
-# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND       #
-# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,     #
-# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,   #
-# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.          #
-# ------------------------------------------------------------------------------------------------ #
-
 import os
 import time
 import logging
 from importlib import reload, import_module
 from types import ModuleType
-from typing import NamedTuple
 
 import jax.numpy as jnp
 import numpy as onp
@@ -657,7 +635,8 @@ def is_stochastic(obj):
     from .._core.stochastic_reward_function import StochasticRewardFunction
     from .._core.stochastic_transition_model import StochasticTransitionModel
     return isinstance(obj, (
-        Policy, StochasticV, StochasticQ, StochasticRewardFunction, StochasticTransitionModel))
+        Policy, StochasticV, StochasticQ, StochasticRewardFunction,
+        StochasticTransitionModel))
 
 
 def is_policy(obj):
@@ -719,7 +698,7 @@ def pretty_repr(o, d=0):
         items = zip(('index', 'data'), (o.index, o.values))
         body = sep + sep.join(f"{k}={pretty_repr(v, d + 1)}" for k, v in items)
         return f"{type(o).__name__}({body})"
-    if isinstance(o, NamedTuple) or hasattr(o, '_asdict'):
+    if hasattr(o, '_asdict'):
         sep = '\n' + i * (d + 1)
         body = sep + sep.join(f"{k}={pretty_repr(v, d + 1)}" for k, v in o._asdict().items())
         return f"{type(o).__name__}({body})"
